import 'package:flutter/material.dart';
import 'dart:async';

class MillNotify extends StatefulWidget {
  final String text;
  final TextStyle textStyle;
  final double height;
  final Color color;
  final Widget prefixIcon;
  final Widget suffixIcon;

  ///滚动方向，水平或者垂直
  final Axis scrollAxis;

  ///空白部分占控件的百分比
  final double ratioOfBlankToScreen;

  MillNotify({
    Key key,
    @required this.text,
    this.height = 38,
    this.color,
    this.textStyle,
    this.scrollAxis: Axis.horizontal,
    this.ratioOfBlankToScreen: 0.25,
    this.prefixIcon,
    this.suffixIcon,
  })  : assert(
          text != null,
        ),
        super(key: key);

  @override
  State<StatefulWidget> createState() {
    return new MillNotifyState();
  }
}

class MillNotifyState extends State<MillNotify>
    with SingleTickerProviderStateMixin {
  ScrollController scroController;
  double blankWidth;
  double blankHeight;
  double position = 0.0;
  Timer timer;
  final double _moveDistance = 3.0;
  final int _timerRest = 100;
  GlobalKey _key = GlobalKey();

  @override
  void initState() {
    super.initState();
    scroController = new ScrollController();
    WidgetsBinding.instance.addPostFrameCallback((callback) {
      startTimer();
    });
  }

  void startTimer() {
    double widgetWidth =
        _key.currentContext.findRenderObject().paintBounds.size.width;
    double widgetHeight =
        _key.currentContext.findRenderObject().paintBounds.size.height;

    timer = Timer.periodic(new Duration(milliseconds: _timerRest), (timer) {
      double maxScrollExtent = scroController.position.maxScrollExtent;
      double pixels = scroController.position.pixels;
      //当animateTo的距离大于最大滑动距离时，则要返回第一个child的特定位置，让末尾正好处于最右侧，然后继续滚动，造成跑马灯的假象
      if (pixels + _moveDistance >= maxScrollExtent) {
        if (widget.scrollAxis == Axis.horizontal) {
          position = (maxScrollExtent - blankWidth - widgetWidth) / 2 +
              pixels -
              maxScrollExtent;
        } else {
          position = (maxScrollExtent - blankHeight - widgetHeight) / 2 +
              pixels -
              maxScrollExtent;
        }
        scroController.jumpTo(position);
      }
      position += _moveDistance;
      scroController.animateTo(position,
          duration: new Duration(milliseconds: _timerRest),
          curve: Curves.linear);
    });
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    double screenWidth = MediaQuery.of(context).size.width;
    double screenHeight = MediaQuery.of(context).size.height;
    blankWidth = screenWidth * widget.ratioOfBlankToScreen;
    blankHeight = screenHeight * widget.ratioOfBlankToScreen;
  }

  Widget getBothEndsChild() {
    if (widget.scrollAxis == Axis.vertical) {
      String newString = widget.text.split("").join("\n");
      return new Center(
        child: new Text(
          newString,
          style: widget.textStyle,
          textAlign: TextAlign.center,
        ),
      );
    }
    return new Center(
        child: new Text(
      widget.text,
      style: widget.textStyle,
    ));
  }

  Widget getCenterChild() {
    if (widget.scrollAxis == Axis.horizontal) {
      return new Container(width: blankWidth);
    } else {
      return new Container(height: blankHeight);
    }
  }

  @override
  void dispose() {
    scroController.dispose();
    super.dispose();
    timer.cancel();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      height: widget.height,
      color: widget.color,
      padding: EdgeInsets.only(left: 10, right: 10),
      child: Row(
        children: <Widget>[
          widget.prefixIcon != null
              ? Container(
                  child: widget.prefixIcon,
                  padding: EdgeInsets.only(right: 8),
                )
              : Container(),
          Expanded(
            flex: 1,
            child: new ListView(
              key: _key,
              scrollDirection: widget.scrollAxis,
              controller: scroController,
              physics: new NeverScrollableScrollPhysics(),
              children: <Widget>[
                getBothEndsChild(),
                getCenterChild(),
                getBothEndsChild(),
              ],
            ),
          ),
          widget.suffixIcon != null
              ? Container(
                  child: widget.suffixIcon,
                  padding: EdgeInsets.only(left: 8),
                )
              : Container(),
        ],
      ),
    );
  }
}
