import 'package:flutter/scheduler.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'dart:math';

class MillLoadWidget extends SingleChildRenderObjectWidget {
  const MillLoadWidget({
    Key key,
    Widget child,
    this.hasLayoutExtent = false,
    this.loadIndicatorLayoutExtent = 0.0,
    @required this.extraExtentNotifier,
  }) : super(key: key, child: child);

  final bool hasLayoutExtent;
  final double loadIndicatorLayoutExtent;
  final ValueNotifier<double> extraExtentNotifier;

  @override
  MillLoadSliver createRenderObject(BuildContext context) {
    return MillLoadSliver(
      hasLayoutExtent: hasLayoutExtent,
      loadIndicatorExtent: loadIndicatorLayoutExtent,
      extraExtentNotifier: extraExtentNotifier,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, covariant MillLoadSliver renderObject) {
    renderObject
      ..hasLayoutExtent = hasLayoutExtent
      ..loadIndicatorLayoutExtent = loadIndicatorLayoutExtent;
  }
}

class MillLoadSliver extends RenderSliverSingleBoxAdapter {
  MillLoadSliver({
    RenderBox child,
    @required bool hasLayoutExtent,
    @required double loadIndicatorExtent,
    @required this.extraExtentNotifier,
  })  : _hasLayoutExtent = hasLayoutExtent,
        _loadIndicatorExtent = loadIndicatorExtent,
        super(child: child);

  final ValueNotifier<double> extraExtentNotifier;

  double _loadIndicatorExtent;

  double get loadIndicatorLayoutExtent => _loadIndicatorExtent;

  set loadIndicatorLayoutExtent(double value) {
    assert(value != null);
    assert(value >= 0.0);
    if (value == _loadIndicatorExtent) return;
    _loadIndicatorExtent = value;
    markNeedsLayout();
  }

  bool get hasLayoutExtent => _hasLayoutExtent;
  bool _hasLayoutExtent;

  set hasLayoutExtent(bool value) {
    assert(value != null);
    if (value == _hasLayoutExtent) return;
    _hasLayoutExtent = value;
    markNeedsLayout();
  }

  double get childSize =>
      constraints.axis == Axis.vertical ? child.size.height : child.size.width;

  double layoutExtentOffsetCompensation = 0.0;

  @override
  void performLayout() {
    double extraExtent = 0.0;
    if (constraints.precedingScrollExtent <
        constraints.viewportMainAxisExtent) {
      extraExtent = constraints.viewportMainAxisExtent -
          constraints.precedingScrollExtent;
    }
    extraExtentNotifier.value = extraExtent;
    final double layoutExtent = _hasLayoutExtent ? 60.0 : 0.0;
    final bool active =
        constraints.remainingPaintExtent > 1.0 || layoutExtent >= 0.0;

    final double overscrolledExtent = max(
        constraints.remainingPaintExtent +
            (constraints.precedingScrollExtent < _loadIndicatorExtent
                ? constraints.scrollOffset
                : 0.0),
        0.0);
    child.layout(
      constraints.asBoxConstraints(
        maxExtent: _hasLayoutExtent
            ? max(_loadIndicatorExtent, overscrolledExtent)
            : overscrolledExtent,
      ),
      parentUsesSize: true,
    );

    if (active) {
      geometry = SliverGeometry(
        scrollExtent: layoutExtent,
        paintOrigin: -constraints.scrollOffset,
        paintExtent: max(
          // Check child size (which can come from overscroll) because
          // layoutExtent may be zero. Check layoutExtent also since even
          // with a layoutExtent, the indicator builder may decide to not
          // build anything.
          min(max(childSize, layoutExtent), constraints.remainingPaintExtent) -
              constraints.scrollOffset,
          0.0,
        ),
        maxPaintExtent: max(
          min(max(childSize, layoutExtent), constraints.remainingPaintExtent) -
              constraints.scrollOffset,
          0.0,
        ),
        layoutExtent: min(max(layoutExtent - constraints.scrollOffset, 0.0),
            constraints.remainingPaintExtent),
      );
    } else {
      // If we never started overscrolling, return no geometry.
      geometry = SliverGeometry.zero;
    }
  }

  @override
  void paint(PaintingContext paintContext, Offset offset) {
    if (constraints.remainingPaintExtent > 0.0 ||
        constraints.scrollOffset + childSize > 0) {
      paintContext.paintChild(child, offset);
    }
  }

  @override
  void applyPaintTransform(RenderObject child, Matrix4 transform) {}
}

enum LoadMode {
  /// Initial state, when not being overscrolled into, or after the overscroll
  /// is canceled or after done and the sliver retracted away.
  inactive,

  /// While being overscrolled but not far enough yet to trigger the refresh.
  drag,

  /// While the onLoad task is running.
  load,

  /// 刷新完成
  loaded,

  /// While the indicator is animating away after refreshing.
  done,
}

class SliveFooter extends StatefulWidget {
  SliveFooter({
    this.onLoad,
    this.focusNotifier,
    this.loading,
    this.backgroundColor,
    this.extraExtentNotifier,
    this.hasMore,
    this.footReleaseText,
    this.noDataText,
  });
  final ValueNotifier<bool> focusNotifier;
  final Widget loading;
  final Function onLoad;
  final Color backgroundColor;
  final ValueNotifier<double> extraExtentNotifier;
  final bool hasMore;
  final String footReleaseText;
  final String noDataText;

  _SliveFooter createState() => _SliveFooter();
}

class _SliveFooter extends State<SliveFooter> {
  bool _hasSliverLayoutExtent = false;
  double _latestIndicatorBoxExtent = 0.0;
  double _refreshTriggerPullDistance = 60.0;
  LoadMode loadState;

  bool get _focus => widget.focusNotifier.value;

  static const double _inactiveResetOverscrollFraction = 0.1;

  Future<void> loadTask;

  // 初始化
  @override
  void initState() {
    super.initState();
    loadState = LoadMode.inactive;
  }

  LoadMode _transitionNextState() {
    // 完成
    LoadMode goToFinish() {
      LoadMode state = LoadMode.loaded;
      Future.delayed(Duration(milliseconds: 800), () {
        if (mounted) setState(() => _hasSliverLayoutExtent = false);
      });
      return state;
    }

    LoadMode nextState;
    switch (loadState) {
      case LoadMode.inactive:
        if (_latestIndicatorBoxExtent <= 0 || !_focus) {
          return LoadMode.inactive;
        } else {
          nextState = LoadMode.drag;
        }
        continue drag;
      drag:
      case LoadMode.drag:
        if (_latestIndicatorBoxExtent == 0) {
          return LoadMode.inactive;
        } else if (_latestIndicatorBoxExtent <= _refreshTriggerPullDistance) {
          if (_hasSliverLayoutExtent) {
            SchedulerBinding.instance
                .addPostFrameCallback((Duration timestamp) {
              if (mounted) setState(() => _hasSliverLayoutExtent = false);
            });
          }
          return LoadMode.drag;
        } else {
          // 提前固定高度，防止列表回弹
          SchedulerBinding.instance.addPostFrameCallback((Duration timestamp) {
            if (!_hasSliverLayoutExtent) {
              if (mounted) setState(() => _hasSliverLayoutExtent = true);
            }
          });
          if (widget.onLoad != null) {
            if (!_focus) {
              SchedulerBinding.instance
                  .addPostFrameCallback((Duration timestamp) {
                loadTask = widget.onLoad()
                  ..then((_) {
                    if (mounted) {
                      setState(() => loadTask = null);
                      loadState = _transitionNextState();
                    }
                  });
              });
              return LoadMode.load;
            }
          }
          return LoadMode.drag;
        }
        // Don't continue here. We can never possibly call onLoad and
        // progress to the next state in one [computeNextState] call.
        break;
      case LoadMode.load:
        if (loadTask != null) {
          return LoadMode.load;
        } else {
          // 结束
          var state = goToFinish();
          if (state != null) return state;
        }
        continue done;
      done:
      case LoadMode.done:
        if (_latestIndicatorBoxExtent >
            _refreshTriggerPullDistance * _inactiveResetOverscrollFraction) {
          return LoadMode.done;
        } else {
          nextState = LoadMode.inactive;
        }
        break;
      case LoadMode.loaded:
        nextState = loadState;
        if (_latestIndicatorBoxExtent <= 60.0 &&
            !_focus &&
            !_hasSliverLayoutExtent) {
          nextState = LoadMode.inactive;
        }
        break;
      default:
        break;
    }

    return nextState;
  }

  @override
  Widget build(BuildContext context) {
    return MillLoadWidget(
      hasLayoutExtent: _hasSliverLayoutExtent,
      loadIndicatorLayoutExtent: 60,
      extraExtentNotifier: widget.extraExtentNotifier,
      child: LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
          // 列表未占满时恢复一下状态
          if (widget.extraExtentNotifier.value > 0.0 &&
              loadState == LoadMode.loaded &&
              loadTask == null) {
            loadState = LoadMode.inactive;
          }
          _latestIndicatorBoxExtent =
              constraints.maxHeight - widget.extraExtentNotifier.value;
          loadState = _transitionNextState();
          if (widget.hasMore && _latestIndicatorBoxExtent >= 0) {
            return loadState == LoadMode.drag &&
                    _latestIndicatorBoxExtent > _refreshTriggerPullDistance
                ? Container(
                    color: Color(0xfff3f3f3),
                    alignment: Alignment.topCenter,
                    padding: EdgeInsets.only(top: 15),
                    child: Text(widget.footReleaseText ?? '释放加载'),
                  )
                : Container(
                    alignment: Alignment.center,
                    height: _latestIndicatorBoxExtent,
                    color: Color(0xfff3f3f3),
                    child: loadState == LoadMode.load || _hasSliverLayoutExtent
                        ? widget.loading == null
                            ? Text('加载中...')
                            : widget.loading
                        : Container(),
                  );
          } else if (!widget.hasMore && _latestIndicatorBoxExtent >= 0) {
            return Container(
              alignment: Alignment.center,
              height: _latestIndicatorBoxExtent,
              child: Text(widget.noDataText ?? '没有更多数据'),
            );
          } else {
            return Container();
          }
        },
      ),
    );
  }
}
