import 'dart:async';

import 'package:flutter/scheduler.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'dart:math';

class MillRefreshWidget extends SingleChildRenderObjectWidget {
  const MillRefreshWidget({
    Key key,
    Widget child,
    this.hasLayoutExtent = false,
  }) : super(key: key, child: child);

  final bool hasLayoutExtent;

  @override
  MillRefreshSliver createRenderObject(BuildContext context) {
    return MillRefreshSliver(
      hasLayoutExtent: hasLayoutExtent,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, covariant MillRefreshSliver renderObject) {
    renderObject..hasLayoutExtent = hasLayoutExtent;
  }
}

class MillRefreshSliver extends RenderSliverSingleBoxAdapter {
  MillRefreshSliver({
    RenderBox child,
    @required bool hasLayoutExtent,
  })  : _hasLayoutExtent = hasLayoutExtent,
        super(child: child);

  bool get hasLayoutExtent => _hasLayoutExtent;
  bool _hasLayoutExtent;

  set hasLayoutExtent(bool value) {
    assert(value != null);
    if (value == _hasLayoutExtent) return;
    _hasLayoutExtent = value;
    markNeedsLayout();
  }

  double get childSize =>
      constraints.axis == Axis.vertical ? child.size.height : child.size.width;

  double layoutExtentOffsetCompensation = 0.0;

  @override
  void layout(Constraints constraints, {bool parentUsesSize = false}) {
    // Header浮动时保持刷新
    super.layout(constraints, parentUsesSize: parentUsesSize);
  }

  @override
  void performLayout() {
    if (child == null) {
      geometry = SliverGeometry.zero;
      return;
    }

    final double layoutExtent = _hasLayoutExtent ? 60.0 : 0.0;
    if (layoutExtent != layoutExtentOffsetCompensation) {
      geometry = SliverGeometry(
        scrollOffsetCorrection: layoutExtent - layoutExtentOffsetCompensation,
      );
      layoutExtentOffsetCompensation = layoutExtent;
      return;
    }
    final bool active = constraints.overlap < 0.0 || layoutExtent > 0.0;
    final double overscrolledExtent =
        constraints.overlap < 0.0 ? constraints.overlap.abs() : 0.0;
    child.layout(
      constraints.asBoxConstraints(
        maxExtent: layoutExtent + overscrolledExtent,
      ),
      parentUsesSize: true,
    );
    if (active) {
      geometry = SliverGeometry(
        scrollExtent: layoutExtent,
        paintOrigin: -overscrolledExtent - constraints.scrollOffset,
        paintExtent: min(
            max(
              max(childSize, layoutExtent) - constraints.scrollOffset,
              0.0,
            ),
            constraints.remainingPaintExtent),
        maxPaintExtent: max(
          max(childSize, layoutExtent) - constraints.scrollOffset,
          0.0,
        ),
        layoutExtent: max(layoutExtent - constraints.scrollOffset, 0.0),
      );
    } else {
      // If we never started overscrolling, return no geometry.
      geometry = SliverGeometry.zero;
    }
  }

  @override
  void paint(PaintingContext paintContext, Offset offset) {
    if (constraints.overlap < 0.0 || constraints.scrollOffset + childSize > 0) {
      paintContext.paintChild(child, offset);
    }
  }

  @override
  void applyPaintTransform(RenderObject child, Matrix4 transform) {}
}

typedef OnRefreshCallback = Future<void> Function();

enum RefreshMode { inactive, drag, refresh, refreshed, done, error }

class SliveHeader extends StatefulWidget {
  SliveHeader({
    this.onRefresh,
    this.focusNotifier,
    this.loading,
    this.finished,
    this.timeout,
    this.error,
    this.backgroundColor,
    this.dropText,
    this.releaseText,
  });
  final ValueNotifier<bool> focusNotifier;
  final Widget loading;
  final Widget finished;
  final Widget error;
  final int timeout;
  final Function onRefresh;
  final Color backgroundColor;
  final String dropText;
  final String releaseText;

  _SliveHeader createState() => _SliveHeader();
}

class _SliveHeader extends State<SliveHeader> {
  RefreshMode refreshState;
  double _latestIndicatorBoxExtent = 0.0;
  double _refreshTriggerPullDistance = 100.0;
  bool _hasSliverLayoutExtent = false;
  Future<void> _refreshTask;
  bool _error = false;
  static const double _inactiveResetOverscrollFraction = 0.1;

  bool get _focus => widget.focusNotifier.value;

  @override
  void initState() {
    super.initState();
    refreshState = RefreshMode.inactive;
  }

  RefreshMode _transitionNextState() {
    RefreshMode nextState;

    // 完成
    RefreshMode goToFinish() {
      RefreshMode state = RefreshMode.refreshed;
      Future.delayed(Duration(milliseconds: 560), () {
        if (mounted)
          setState(() {
            _hasSliverLayoutExtent = false;
            _error = false;
          });
      });
      return state;
    }

    RefreshMode goToError() {
      RefreshMode state = RefreshMode.error;
      Future.delayed(Duration(milliseconds: 560), () {
        if (mounted) {
          setState(() {
            _hasSliverLayoutExtent = false;
            _error = false;
          });
        }
      });
      return state;
    }

    switch (refreshState) {
      case RefreshMode.inactive:
        if (_latestIndicatorBoxExtent <= 0 || (!_focus)) {
          return RefreshMode.inactive;
        } else {
          nextState = RefreshMode.drag;
        }
        continue drag;
      drag:
      case RefreshMode.drag:
        if (_latestIndicatorBoxExtent == 0) {
          return RefreshMode.inactive;
        } else if (_latestIndicatorBoxExtent <= _refreshTriggerPullDistance) {
          if (_hasSliverLayoutExtent) {
            SchedulerBinding.instance
                .addPostFrameCallback((Duration timestamp) {
              if (mounted) {
                setState(() {
                  _hasSliverLayoutExtent = false;
                  _error = false;
                });
              }
            });
          }
          return RefreshMode.drag;
        } else {
          if (widget.onRefresh != null) {
            SchedulerBinding.instance
                .addPostFrameCallback((Duration timestamp) {
              if (!_hasSliverLayoutExtent) {
                if (mounted) setState(() => _hasSliverLayoutExtent = true);
              }
            });
            if (!_focus) {
              SchedulerBinding.instance
                  .addPostFrameCallback((Duration timestamp) {
                Timer t = Timer(Duration(seconds: widget.timeout), () {
                  if (mounted && _refreshTask != null && !_error) {
                    setState(() {
                      _error = true;
                      _refreshTask = null;
                    });
                    refreshState = _transitionNextState();
                  }
                });
                _refreshTask = widget.onRefresh()
                  ..then((_) {
                    if (t.isActive) {
                      t.cancel();
                    }
                    if (mounted && _refreshTask != null) {
                      setState(() => _refreshTask = null);
                      refreshState = _transitionNextState();
                    }
                  });
              });
              return RefreshMode.refresh;
            }
          }
          return RefreshMode.drag;
        }
        break;
      case RefreshMode.refresh:
        if (_refreshTask != null) {
          return RefreshMode.refresh;
        } else if (_error) {
          var state = goToError();
          if (state != null) {
            return state;
          }
        } else {
          // 完成
          var state = goToFinish();
          if (state != null) {
            return state;
          }
        }
        continue done;
      done:
      case RefreshMode.done:
        if (_latestIndicatorBoxExtent >
            _refreshTriggerPullDistance * _inactiveResetOverscrollFraction) {
          return RefreshMode.done;
        } else {
          nextState = RefreshMode.inactive;
        }
        break;
      case RefreshMode.refreshed:
        nextState = refreshState;
        if (_latestIndicatorBoxExtent <= 0 &&
            !_focus &&
            !_hasSliverLayoutExtent) {
          nextState = RefreshMode.inactive;
        }
        break;
      case RefreshMode.error:
        nextState = refreshState;
        if (_latestIndicatorBoxExtent <= 0 &&
            !_focus &&
            !_hasSliverLayoutExtent) {
          nextState = RefreshMode.inactive;
        }
        break;
      default:
        break;
    }

    return nextState;
  }

  @override
  Widget build(BuildContext context) {
    Widget loading = Container(
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          Container(
            width: 20.0,
            height: 20.0,
            child: CircularProgressIndicator(
              strokeWidth: 2.0,
              valueColor: AlwaysStoppedAnimation(
                Color(0xff333333),
              ),
            ),
          ),
          Padding(padding: EdgeInsets.only(right: 15)),
          Text(
            '加载中...',
            style: TextStyle(
              color: Color(0xff333333),
            ),
          )
        ],
      ),
    );

    Widget finished = Container(
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          Icon(Icons.check, color: Color(0xff409eff)),
          Padding(
            padding: EdgeInsets.only(left: 15),
          ),
          Text(
            '加载完成',
            style: TextStyle(
              color: Color(0xff409eff),
            ),
          ),
        ],
      ),
    );

    Widget error = Container(
      child: Row(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [Text('加载失败')],
      ),
    );

    return MillRefreshWidget(
      hasLayoutExtent: _hasSliverLayoutExtent,
      child: LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
          _latestIndicatorBoxExtent = constraints.maxHeight;
          refreshState = _transitionNextState();
          if (_latestIndicatorBoxExtent <= 15) {
            return Container();
          }

          return Container(
            color: widget.backgroundColor,
            alignment: Alignment.topCenter,
            width: double.infinity,
            height:
                _latestIndicatorBoxExtent > 60 ? _latestIndicatorBoxExtent : 60,
            child: Column(
              mainAxisAlignment: MainAxisAlignment.center,
              children: [
                refreshState == RefreshMode.inactive
                    ? Container()
                    : refreshState != RefreshMode.drag
                        ? Expanded(
                            flex: 1,
                            child: refreshState == RefreshMode.refresh &&
                                        _hasSliverLayoutExtent
                                ? widget.loading == null
                                    ? loading
                                    : widget.loading
                                : refreshState == RefreshMode.error
                                    ? widget.error ?? error
                                    : (refreshState == RefreshMode.refreshed ||
                                        refreshState == RefreshMode.done)
                                        ? widget.finished == null
                                            ? finished
                                            : widget.finished
                                        : Container(),
                          )
                        : Container(),
                refreshState == RefreshMode.drag
                    ? Expanded(
                        child: Container(
                          alignment: Alignment.bottomCenter,
                          padding: EdgeInsets.only(bottom: 20),
                          child: Text(_latestIndicatorBoxExtent <
                                  _refreshTriggerPullDistance
                              ? widget.dropText ?? '下拉刷新'
                              : widget.releaseText ?? '释放加载',
                              style: TextStyle(
                                color: Color(0xff888888),
                              ),
                            ),
                        ),
                      )
                    : Container(),
              ],
            ),
          );
        },
      ),
    );
  }
}
