import 'package:flutter/material.dart';
import 'dart:async';

enum MillSwipeType { scaleCard, card, normal }

class MillSwipe extends StatefulWidget {
  MillSwipe({
    Key key,
    this.controller,
    this.delay = 3000,
    this.pageList = const [],
    this.indicatorColor = const Color(0xffc3c3c3),
    this.indicatorActiveColor = const Color(0xff409eff),
    this.indicatorSize = 14,
    this.indicatorSpace = 4,
    this.type = MillSwipeType.normal,
    this.viewportFraction = 1.0,
    this.space = 5.0,
    this.direction = Axis.horizontal,
    this.autoPlay = false,
    this.height = 130.0,
    this.width = 350.0,
  }) : super(key: key);

  final MillSwiperController controller;
  final int delay;
  final List<Widget> pageList;
  final Color indicatorColor;
  final Color indicatorActiveColor;
  final double indicatorSize;
  final double indicatorSpace;
  final MillSwipeType type;
  final double viewportFraction;
  final double space;
  final Axis direction;
  final bool autoPlay;
  final double height;
  final double width;

  @override
  _MillSwipe createState() => new _MillSwipe();
}

class _MillSwipe extends State<MillSwipe> {
  PageController _pageController;
  MillSwiperController _swiperController;
  var _currPageValue = 2.0;
  double _scaleFactor = .8;
  Timer _timer;

  @override
  void initState() {
    super.initState();
    _swiperController = widget.controller;
    _swiperController ??= MillSwiperController(widget.autoPlay);
    _intAutoPlay();
    _swiperController.addListener(() {
      _intAutoPlay();
      switch (_swiperController.event) {
        case MillSwiperController.NEXT:
          _pageController.nextPage(
            duration: new Duration(milliseconds: 100),
            curve: Curves.linear,
          );
          break;
        case MillSwiperController.PREVIOUS:
          _pageController.previousPage(
            duration: new Duration(milliseconds: 100),
            curve: Curves.linear,
          );
          break;
        case MillSwiperController.SKIP:
          int index = _swiperController.getIndex();
          if (index + 2 != _currPageValue.round()) {
            _pageController.animateToPage(
              index + 2,
              duration: new Duration(milliseconds: 100),
              curve: Curves.linear,
            );
          }
          break;
      }
    });
    double viewportFraction = widget.viewportFraction;
    if (widget.type != MillSwipeType.normal && viewportFraction >= 0.94) {
      viewportFraction = 0.91;
    }
    _pageController = PageController(
      viewportFraction: viewportFraction,
      initialPage: 2,
    );
    _pageController.addListener(() {
      setState(() {
        _currPageValue = _pageController.page;
        int index = _currPageValue.floor();
        if (index != _swiperController.getIndex()) {
          _swiperController.setIndex(index);
        }
        int l = widget.pageList.length;
        if (_currPageValue.ceil() <= 1) {
          _currPageValue = l + 1.0;
          _pageController.jumpToPage(l + 1);
        } else if (index >= l + 2) {
          _currPageValue = 2.0;
          _pageController.jumpToPage(2);
        }
      });
    });
  }

  @override
  void dispose() {
    _stopAutoPlay();
    _swiperController.dispose();
    _pageController.dispose();
    super.dispose();
  }

  _intAutoPlay() {
    bool autoPlay = _swiperController.getAutoPlay();
    if (autoPlay && _timer == null) {
      _startAutoPlay();
    } else if (!autoPlay && _timer != null) {
      _stopAutoPlay();
    }
  }

  _startAutoPlay() {
    _timer = Timer.periodic(
      Duration(milliseconds: widget.delay),
      (Timer timer) {
        return _swiperController.next();
      },
    );
  }

  void _stopAutoPlay() {
    if (_timer != null) {
      _timer.cancel();
      _timer = null;
    }
  }

  @override
  Widget build(BuildContext context) {
    List pages = [];
    int l = widget.pageList.length;
    if (l > 0) {
      pages.add(widget.pageList[l - 2]);
      pages.add(widget.pageList[l - 1]);
      pages.addAll(widget.pageList);
      pages.add(widget.pageList[0]);
      pages.add(widget.pageList[1]);
    }
    return Stack(
      children: [
        PageView.builder(
          scrollDirection: widget.direction,
          itemCount: pages.length,
          controller: _pageController,
          itemBuilder: (context, index) {
            if (widget.type == MillSwipeType.scaleCard) {
              return _buildPageCardItem(index, pages);
            } else if (widget.type == MillSwipeType.card) {
              return _buildPageDefalutItem(index, pages);
            } else {
              return _buildPageItem(index, pages);
            }
          },
        ),
        widget.direction == Axis.horizontal
            ? Positioned(
                bottom: 15,
                left: 0,
                right: 0,
                child: Align(
                  widthFactor: 1,
                  alignment: Alignment.center,
                  child: Row(
                    mainAxisAlignment: MainAxisAlignment.center,
                    children: widget.pageList.map((el) {
                      int index =
                          widget.pageList.indexWhere((ele) => ele == el);
                      return GestureDetector(
                        onTap: () {
                          _swiperController.skipTo(index);
                        },
                        child: Container(
                          padding: EdgeInsets.symmetric(
                              horizontal: widget.indicatorSpace),
                          child: Container(
                            width: widget.indicatorSize,
                            height: widget.indicatorSize,
                            decoration: BoxDecoration(
                              color: _currPageValue.floor() - 2 != index
                                  ? widget.indicatorColor
                                  : widget.indicatorActiveColor,
                              borderRadius: BorderRadius.circular(10),
                            ),
                          ),
                        ),
                      );
                    }).toList(),
                  ),
                ),
              )
            : Positioned(
                bottom: 0,
                top: 0,
                left: 10,
                child: Align(
                  heightFactor: 1,
                  alignment: Alignment.center,
                  child: Column(
                    mainAxisAlignment: MainAxisAlignment.center,
                    children: widget.pageList.map((el) {
                      int index =
                          widget.pageList.indexWhere((ele) => ele == el);
                      return GestureDetector(
                        onTap: () {
                          _swiperController.skipTo(index);
                        },
                        child: Container(
                          padding: EdgeInsets.symmetric(
                              vertical: widget.indicatorSpace),
                          child: Container(
                            width: widget.indicatorSize,
                            height: widget.indicatorSize,
                            decoration: BoxDecoration(
                              color: _currPageValue.floor() - 2 != index
                                  ? widget.indicatorColor
                                  : widget.indicatorActiveColor,
                              borderRadius: BorderRadius.circular(10),
                            ),
                          ),
                        ),
                      );
                    }).toList(),
                  ),
                ),
              ),
      ],
    );
  }

  _buildPageCardItem(int index, List pages) {
    bool ish = widget.direction == Axis.horizontal;
    double h = widget.height;
    if (!ish) {
      h = widget.width;
    }
    Matrix4 matrix4 = Matrix4.identity();
    if (index == _currPageValue.floor()) {
      //当前的item
      var currScale = 1 - (_currPageValue - index) * (1 - _scaleFactor);
      var currTrans = h * (1 - currScale) / 2;
      if (ish) {
        matrix4 = Matrix4.diagonal3Values(1.0, currScale, 1.0)
          ..setTranslationRaw(0.0, currTrans, 0.0);
      } else {
        matrix4 = Matrix4.diagonal3Values(currScale, 1.0, 1.0)
          ..setTranslationRaw(currTrans, 0.0, 0.0);
      }
    } else if (index == _currPageValue.floor() + 1) {
      //右边的item
      var currScale =
          _scaleFactor + (_currPageValue - index + 1) * (1 - _scaleFactor);
      var currTrans = h * (1 - currScale) / 2;
      if (ish) {
        matrix4 = Matrix4.diagonal3Values(1.0, currScale, 1.0)
          ..setTranslationRaw(0.0, currTrans, 0.0);
      } else {
        matrix4 = Matrix4.diagonal3Values(currScale, 1.0, 1.0)
          ..setTranslationRaw(currTrans, 0.0, 0.0);
      }
    } else if (index == _currPageValue.floor() - 1) {
      //左边
      var currScale = 1 - (_currPageValue - index) * (1 - _scaleFactor);
      var currTrans = h * (1 - currScale) / 2;
      if (ish) {
        matrix4 = Matrix4.diagonal3Values(1.0, currScale, 1.0)
          ..setTranslationRaw(0.0, currTrans, 0.0);
      } else {
        matrix4 = Matrix4.diagonal3Values(currScale, 1.0, 1.0)
          ..setTranslationRaw(currTrans, 0.0, 0.0);
      }
    } else {
      //其他，不在屏幕显示的item
      matrix4 = Matrix4.diagonal3Values(1.0, _scaleFactor, 1.0)
        ..setTranslationRaw(0.0, h * (1 - _scaleFactor) / 2, 0.0);
    }

    return Transform(
      transform: matrix4,
      child: Padding(
        padding: ish
            ? EdgeInsets.symmetric(horizontal: widget.space)
            : EdgeInsets.symmetric(vertical: widget.space),
        child: _buildPageItem(index, pages),
      ),
    );
  }

  _buildPageDefalutItem(int index, List pages) {
    bool ish = widget.direction == Axis.horizontal;

    return Padding(
      padding: ish
          ? EdgeInsets.symmetric(horizontal: widget.space)
          : EdgeInsets.symmetric(vertical: widget.space),
      child: _buildPageItem(index, pages),
    );
  }

  _buildPageItem(int index, List pages) {
    return Container(
      child: pages[index],
    );
  }
}

class MillSwiperController extends ChangeNotifier {
  MillSwiperController(this.autoPlay);

  /// Previous
  static const int PREVIOUS = -1;

  /// Next
  static const int NEXT = 1;

  static const int SKIP = 2;

  Completer _completer;
  bool autoPlay = false;

  int event;
  int currentIndex;

  setIndex(index) {
    this.currentIndex = index;
  }

  int getIndex() {
    return this.currentIndex;
  }

  bool getAutoPlay() {
    return this.autoPlay;
  }

  void startAutoPlay() {
    this.event = 0;
    this.autoPlay = true;
    notifyListeners();
  }

  void stopAutoPlay() {
    this.event = 0;
    this.autoPlay = false;
    notifyListeners();
  }

  Future next() {
    this.event = NEXT;
    _completer = Completer();
    notifyListeners();
    return _completer.future;
  }

  Future previous() {
    this.event = PREVIOUS;
    _completer = Completer();
    notifyListeners();
    return _completer.future;
  }

  Future skipTo(int index) {
    this.event = SKIP;
    this.setIndex(index);
    _completer = Completer();
    notifyListeners();
    return _completer.future;
  }

  void complete() {
    if (!_completer.isCompleted) {
      _completer.complete();
    }
  }
}
